
Files in this directory describe a sample run of Autoclass, applied
to a database of Imported Car Models sold in 1985.  Although the results
from this run aren't particularly spectacular, it has been chosen because
most readers will have some intuition for the subject, and know what most
of the attributes used mean.

The procedure by which these files were created is described in: 

    scriptc.text

A direct dump of the screen from the run is in: 

    screenc.text

Raw Input:

    imports-85.names
    imports-85c.db2 (renamed from imports-85.data)

Input descriptions processed for Autoclass:

    imports-85c.hd2
    imports-85c.model
    imports-85c.s-params

Outputs from the direct run:

    imports-85c.log
    imports-85c.search
    ;; imports-85c.results-bin not included to save space

Outputs from post processing of the results file:

    imports-85c.rlog
    imports-85c.influ-o-text-1
    imports-85c.case-text-1
    imports-85c.class-text-1

    imports-85c.influ-o-data-1
    imports-85c.case-data-1
    imports-85c.class-data-1


